# This is the lower level deamon that is used to do
# all the dirty jobs and make it look tidy to the
# people above .. sigh . well see the program for yourself

import os
import sys
import time
import random
import pickle

import ga2.gaUtils.AgentClass as AgentClass
GA_UTIL_DIR = os.path.dirname(os.path.realpath(__file__))


# current functions included are as follows:
#
# ** uses lock
# *
# 
# 1 ) ** getNewSessionID()
# 2 ) isLocked()
# 3 ) lock()
# 4 ) unlock()
# 5 ) getCurrGen()
# 6 ) setCuttGen()
# 7 ) ** storeAgent()
# 8 ) generateAgentID()
# 9 ) createAgent()
# 10) ** deleteAgent()
# 11) getAgent()
# 12) ** setSession()
# 13) getSession()
# 14) asexualRep()
# 15) mutateAgentObj()


def mutateAgentObj(agentObj, sess):
    """
    INPUT       : sessID
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """

    currDna = agentObj.dna
    itr = 0
    #change 'and' to 'or' and added itr+=1
    while((sess.mutation > random.random()) and (itr < int(len(currDna)/3))):

        if(0.5 > random.random()):
            mutateVal = (sess.mutation)*(10**(sess.spread))
        else:
            mutateVal = -1 * (sess.mutation)*(10**(sess.spread))
        itr+=1
        currDna[random.randint(0, len(currDna)-1)] += mutateVal
        
    return(currDna)
    

def aSexualRep(sess, agentObj):
    """
    INPUT       : agentObj
    OUTPUT      : child agentObj

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """

    childDna = []
    
    itr = 0
    while(itr < len(agentObj.dna)):
        if(sess.genecopy < random.random()):
            childDna.append(random.randint(0, 10**(sess.spread)))
        else:
            childDna.append(agentObj.dna[itr])
        itr += 1
            
    childObj = createAgent(sess)
    childObj.dna = childDna
    #childObj.fitness = agentObj.fitness
    childObj.dna = mutateAgentObj(childObj, sess)
    return(childObj)
        

def getSession(sessID):
    """
    INPUT       : sessID
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    try:
    
        cgfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessID)
                    +"/smtf/SESS.smtf", "rb")
        currImg = pickle.load(cgfp)
        cgfp.close()
    
    except Exception:
        print("error in getting the session {}".format(sessID))
        currImg = None
        
    print(currImg.agentCount)
    return(currImg)


def setSession(sessionObj):
    """
    INPUT       : sessionObj
    OUTPUT      : 1 on success 0 on failure

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    
    lock(sessionObj.sessID)
    #try:
    tpfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessionObj.sessID)
                +"/smtf/SESS.smtf", "wb")
    pickle.dump(sessionObj, tpfp)
    tpfp.close()
    '''
    except Exception:
        print("error in store of session, couldnt wb")
        return(0)
    '''
    unlock(sessionObj.sessID)
    
    return(1)


def getNewSessionID():
    """
    INPUT       : None
    OUTPUT      : New SessionID

    DESCRIPTION : Returns a new SessionID , or 0 when its the
                  first session
    """

    lock()
    try:
        
        sessFp = open(GA_UTIL_DIR+"/utilFiles/SESSIONNEXT.smtf", "r")
        sessID = int(sessFp.readline())
        sessFp.close()
    
    except Exception:
        print("Generating first session")
        sessID = 0
    
    try:
        sessFp = open(GA_UTIL_DIR+"/utilFiles/SESSIONNEXT.smtf", "w")
        sessFp.write(str(sessID+1))
        sessFp.close()

    except Exception:
        print("Error updating the SESSIONNEXT.smtf file")

    unlock()
    lock(sessID)
    unlock(sessID)
    return(sessID)


def isLocked(sessID=None):
    """
    INPUT       : sessID = None
    OUTPUT      : current status of the lock based on 
                  sessIDx

    DESCRIPTION : checks the locks, based on sessID
    """

    if(sessID is not None):
        smtf = "tmp" + str(sessID) + "/LOCK.smtf"
    else:
        smtf = "SESSIONLOCK.smtf" 
    
    try:
        if(not os.path.isdir(GA_UTIL_DIR+"/utilFiles")):
            os.mkdir(GA_UTIL_DIR+"/utilFiles")
        lckfp = open(GA_UTIL_DIR+"/utilFiles/"+smtf, "r")
        lckpos = int(lckfp.readline())
        lckfp.close()

    except Exception:
        lckpos = 0

    return(lckpos)


def lock(sessID=None):
    """
    INPUT       : sessID = None
    OUTPUT      : returns 1 if locked, 0 on error

    DESCRIPTION : locks the session files if sessID is passed
                  else , does it for the entire session
    """

    while(isLocked(sessID)):
        time.sleep(0.5)

    if(sessID is not None):
        smtf = "tmp" + str(sessID) + "/smtf/LOCK.smtf"
    else:
        smtf = "SESSIONLOCK.smtf" 
        
    try:
        lckfp = open(GA_UTIL_DIR+"/utilFiles/"+smtf, "w")
        lckfp.close()
    
    except Exception:
        os.mkdir(GA_UTIL_DIR+"/utilFiles/"+"tmp"+str(sessID))
        os.mkdir(GA_UTIL_DIR+"/utilFiles/"+"tmp"+str(sessID)+"/smtf")
        os.mkdir(GA_UTIL_DIR+"/utilFiles/"+"tmp"+str(sessID)+"/dnaPool")
        
    lckval = 1
    lckfp = open(GA_UTIL_DIR+"/utilFiles/"+smtf, "w")
    lckfp.write(str(lckval))
    lckfp.close()
    return(lckval)


def unlock(sessID=None):
    """
    INPUT       : sessID = None
    OUTPUT      : 1 on successful unlocking 0 on unsuccessful

    DESCRIPTION : unlocks the session files if sessID is passed
                  else , does it for the entire session
    """

    if(sessID is not None):
        smtf = "tmp" + str(sessID) + "/smtf/LOCK.smtf"
    else:
        smtf = "SESSIONLOCK.smtf"

    try:
        lckfp = open(GA_UTIL_DIR+"/utilFiles/"+smtf, "w")
        lckfp.write(str(0))
        lckfp.close()
        
    except Exception:
        print("Couldnt unlock file !")
        return(0)
    
    return(1)


def getCurrGen(sessID):
    """
    INPUT       : sessID
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    try:
        cgfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessID)+"/smtf/VALID.smtf", "rb")
        currImg = pickle.load(cgfp)
        cgfp.close()
    except Exception:
        #print("error in getting the curr gen")
        currImg = set()

    return(currImg)


def setCurrGen(sessID, agentList):
    """
    INPUT       : None
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    try:
        cgfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessID)+"/smtf/VALID.smtf", "wb")
        pickle.dump(agentList, cgfp)
        cgfp.close()
    except Exception:
        print("error in setting curr gen")
        return(0)
    return(1)


def storeAgent(agentObj):
    """
    INPUT       : None
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    
    currAgents = getCurrGen(agentObj.sessID)
    lock(agentObj.sessID)
    try:
        tpfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(agentObj.sessID)+"/dnaPool/dna"
                    +str(agentObj.agentID)+".dna", "wb")
        pickle.dump(agentObj, tpfp)
        tpfp.close()
        currAgents.add(agentObj.agentID)
    except Exception:
        print("error in store agent, couldnt wb")
        return(0)
    
    setCurrGen(agentObj.sessID, currAgents)
    unlock(agentObj.sessID)
    return(agentObj.agentID)


def getAgent(sessID, agentID):
    """
    INPUT       : None
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """

    try:
        tpfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessID)+"/dnaPool/dna"
                    +str(agentID)+".dna", "rb")
        agentObj = pickle.load(tpfp)
        tpfp.close()
    
    except Exception:
        print("Error couldnt retrieve error in getAgent()")
        agentObj = None
    
    return(agentObj)


def generateAgentID(sessID):
    """
    INPUT       : None
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    try:
        if(not os.path.isdir(GA_UTIL_DIR+"/utilFiles")):
            os.mkdir(GA_UTIL_DIR+"/utilFiles")
        nextfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessID)+"/smtf/NEXTID.smtf", "r")
        Idp = int(nextfp.readline())
        nextfp.close()
    except Exception:
        Idp = 0

    try:
        nextfp = open(GA_UTIL_DIR+"/utilFiles/tmp"+str(sessID)+"/smtf/NEXTID.smtf", "w")
        nextfp.write(str(Idp+1))
        nextfp.close()
        return(Idp)

    except Exception:
        print("error in generate agent ID, couldnt write")
        return(-1)


def createAgent(sess):
    """
    INPUT       : None
    OUTPUT      : Agent object

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    
    agent = AgentClass.AgentDna(sess.numParam, sess.spread)
    agent.sessID = sess.sessID
    agent.agentID = generateAgentID(sess.sessID)

    storeAgent(agent)
    return(agent)

def deleteAgent(sessID, agentID):
    """
    INPUT       : sessID, agentID
    OUTPUT      : None

    DESCRIPTION : Creates and stores the newly created AgentDna
                  object in that particular session's directory
    """
    
    lock(sessID)
    try:
        currGen = getCurrGen(sessID)
        if(agentID in currGen):
            toDelFile = GA_UTIL_DIR + ("/utilFiles/tmp" + str(sessID)
                                   + "/dnaPool/dna" + str(agentID)
                                   +".dna")
            os.remove(toDelFile)
            currGen.remove(agentID)
            setCurrGen(sessID, currGen)
            
        else:
            print("Trying to delete an agent not present in the session")

    except Exception:
        print("delete Agent encountered an error")
        return(0)
    
    finally:
        unlock(sessID)
        
    return(1)



from ga2.gaUtils.gaDeamon import *
from . import gaDisc 
import os,signal

def aexit(a,b):
	os.system('rm -Rf '+os.path.dirname(os.path.realpath(__file__))+'/gaUtils/utilFiles/*')
	os.sys.exit()

signal.signal(signal.SIGTERM,aexit)
signal.signal(signal.SIGINT,aexit)
signal.signal(signal.SIGABRT,aexit)

version = 2.8



# This is the DnaClass / AgentClass file all agents
# are objects of this class
import random


def generateDna(numParam, spread):
    """
    INPUT       : rows = 10, columns = 10, spread = 2
    OUTPUT      : class Dna.dna structure

    DESCRIPTION : returns a 2D list whose number of rows is as specified
                  and whose columns is as specified and each value in the
                  matrix is in [0, 10^spread]
    """
    retDna = [random.randint(0, 10**spread) for row in range(numParam)]
    return(retDna)


class AgentDna:
    """
    This is the AgentDna class, all the agents will be
    of this class, this class impliments some of the 
    basic methods to interact with the Agent and also holds
    all the attributes required to identify and measure the
    agent's fitness
    """
    def __init__(self, numParam,  spread):
        
        self.agentID = None
        self.sessID = None
        self.dna = generateDna(numParam=numParam, spread=spread)
        self.fitness = 0



# This program impliments the discrete genetic algorithm
# here, Discrete means that the new generation is generated
# all at the same time by considering the top agents at that
# given point.

import os
import shutil
import random
import ga2.gaUtils.gaDeamon as deamon

class Session:
    """
    This is the session class which is used to access all
    the agents in the current session, invoke different
    functions on this session and setting session
    parameters.
    """

    def __init__(self,sessID=None, agentCount=100, numParam=10, fixParam=True,
                 spread=2, survival=0.05, genecopy=0.9, mutation=0.01, generateMode='A'):

        if(sessID is None):
            self.sessID = deamon.getNewSessionID()
            self.agentCount = agentCount
            self.numParam = numParam
            self.fixParam = fixParam
            self.spread = spread
            self.generateMode = generateMode
            self.mutation = mutation
            self.genecopy = genecopy
            self.survival = survival
            deamon.setSession(self)

        else:
            tmpObj = deamon.getSession(sessID)
            self.sessID = tmpObj.sessID
            self.agentCount = tmpObj.agentCount
            self.numParam = tmpObj.numParam
            self.fixParam = tmpObj.fixParam
            self.spread = tmpObj.spread
            self.mutation = mutation
            self.survival = survival
            self.genecopy = genecopy
            self.generateMode = tmpObj.generateMode
        
    def init(self):
        """
        INPUT       : None
        OUTPUT      : Returns 1 on success, None on failure

        DESCRIPTION : Initializes the session with initial
                      values based on the session conf .

        we create self.agentCount number of agents here
        using the deamon function createAgent
        """
        
        try:
            for i in range(self.agentCount):
                agentObj = deamon.createAgent(self)
        
        except Exception as e:
            print("gaDiscrete.init failed with error as : {}".format(e))
        
    def getAllAgents(self):
        """
        INPUT       : None
        OUTPUT      : Returns all the current AgentIDs of all the active
                      agents in a list, or None on error

        DESCRIPTION : Self explanatory from OUTPUT

        
        toRet is the set containing the agentIDs of all
        agents who are currently active.
        toRet is populated by using the deamon function
        getCurrGen
        """
        
        try:
            toRet = deamon.getCurrGen(self.sessID)
        except Exception as e:
            print("gaDiscrete.getAllAgents failed with error as : {}".format(e))
            toRet = None
            
        return(toRet)

    def getAgent(self, agentID):
        """
        INPUT       : AgentID
        OUTPUT      : Returns the agent with the given AgentId, 
                      None on error

        DESCRIPTION : Self explanatory from OUTPUT

        agentObj is the agent corresponding to the agentID = agentID
        and belonging to the session with sessID = sessID
        uses the deamon function
        """

        try:
            agentObj = deamon.getAgent(self.sessID, agentID)
            return(agentObj)
        
        except Exception as e:
            print("gaDiscrete.getAgent failed with error as : {}".format(e))
        
    def updateAgent(self, agentObj):
        """
        INPUT       : Agent object
        OUTPUT      : Returns 1 on success, None on failure

        DESCRIPTION : Updates the status of Agent in the cluster.

        Stores the agent in the current session agent's file
        this storage is for persistance and to save the RAM
        for tree building or other more important jobs
        uses the deamon storeAgent function
        """

        try:
            return(deamon.storeAgent(agentObj))
        
        except Exception as e:
            print("gaDiscrete.updateAgent failed with error as : {}".format(e))            
    
    def getBestAgent(self):
        """
        INPUT       : None
        OUTPUT      : Returns the Agent with the best fitness value
                      at present , and None on error

        DESCRIPTION : Self explanatory from OUTPUT

        This is used to retrieve the best agent in the current
        pool of agents based on their fitness value with
        higher fitness corresponding to being better
        uses deamon functions getCurrGen, getAgent
        """

        try:
            bestAgent = None
            currGen = deamon.getCurrGen(self.sessID)
            for agentID in currGen:
                agent = deamon.getAgent(self.sessID, agentID)
                if((bestAgent is None) or (bestAgent.fitness < agent.fitness)):
                    bestAgent = agent
                    
        except Exception as e:
            print("gaDiscrete.getBestAgent failed with error as : {}".format(e))
            bestAgent = None
        
        return(bestAgent)

    def getAverageFitness(self):
        """
        INPUT       : None
        OUTPUT      : Returns the current avegare of the fitness value
                      of all the agents in the session.

        DESCRIPTION : Self explanatory from OUTPUT

        This function is used to get the average fitness value of
        agents in the current session, this uses deamon functions
        getCurrGen, getAgent
        """

        try:
            totalFitness = 0
            currGen = deamon.getCurrGen(self.sessID)
            for agentID in currGen:
                totalFitness += deamon.getAgent(self.sessID, agentID).fitness

        except Exception:
            print("gaDiscrete.getAverageFitness failed with error as : {}".format(e))
            totalFitness = 0

        return(totalFitness/self.agentCount)
                
    def createNextGen(self):
        """
        INPUT       : None
        OUTPUT      : Returns 1 on success, None on failure

        DESCRIPTION : Creates the next generation based on the session
                      conf that is set in the beginning of the session

        This is the main attraction in this session file XD
        the algorithm follows as this :
        
        a 'numberOfSurvivors' number is selected randomly between
        5 - 10 % of the population size. These many top agents will
        qualify for next generation.
        
        now, 1 % of the agents from the unselected agents pool is selected
        in random and made to come into the survivor agents group as 
        wildcard entry
        
        all the unselected agents are deleted (killed ?)
        now the current population is used to create new children
        by calling a deamon function aSexualRep, which returns a
        new agentObj as a child when a parent agentObj is sent
        this new child-agentObj will inherit the fitness value and
        slightly modified from its parent
        """
        
        try:
            if(self.survival > 0.95):
                survivalPerc = 0.95
                
            numberOfSurvivors = int((random.uniform(self.survival, self.survival+0.05) * self.agentCount))                      
            if(numberOfSurvivors < 1):
                numberOfSurvivors = 1

            listOfAgents = [(agent.agentID, agent.fitness) for agent in
                            [deamon.getAgent(self.sessID, agentID) for agentID in
                             deamon.getCurrGen(self.sessID)]]

            listOfAgents.sort(key=lambda x: x[1], reverse=True)
            unselected = listOfAgents[numberOfSurvivors:]
            wildCardEntries = random.sample(unselected, int(len(unselected)*0.01))
            finalUnselected = [agent for agent in unselected if(agent not in wildCardEntries)]

            for agent in finalUnselected:
                deamon.deleteAgent(self.sessID, agent[0])

            currPop = list(deamon.getCurrGen(self.sessID))
            itr = 0
            while(len(deamon.getCurrGen(self.sessID)) < self.agentCount):
                self.updateAgent(deamon.aSexualRep(self, self.getAgent(currPop[itr])))
                itr = (itr + 1)%(len(currPop))

        except Exception as e:
            print("gaDiscrete.createNextGen failed with error as : {}".format(e))
            
    def resetEnv(self):
        """
        INPUT       : None
        OUTPUT      : Returns 1 on success, None on failure

        DESCRIPTION : Deletes all the agents in the current session
        
        This function effectively resets the session to the
        "Just created form" all the meta data files along with previous
        agents are deleted (killed ?). Uses the deamon (pun intended)
        functions setSession which registeres this session for future use
        """
        
        try:            
            sessDir = deamon.GA_UTIL_DIR + "/utilFiles/tmp" + str(self.sessID)
            shutil.rmtree(sessDir+"/smtf")
            shutil.rmtree(sessDir+"/dnaPool")
            os.mkdir(sessDir+"/smtf")
            os.mkdir(sessDir+"/dnaPool")
            fp = open(sessDir+"/smtf/LOCK.smtf", "w")
            fp.write(str(0))
            fp.close()
            deamon.setSession(self)
            
        except Exception as e:
            print("gaDiscrete.resetEnv failed with error as : {}".format(e))
            

    def delete(self):
        """
        INPUT       : None
        OUTPUT      : Returns 1 on success, 0 on failure

        DESCRIPTION : Closes and deletes the current session
        
        As the name of the function says, this function deletes all of
        the meta data generated by this session and effectively deletes
        this session uses the deamon data GA_UTIL_DIR
        """

        try:
            sessDir = deamon.GA_UTIL_DIR + "/utilFiles/tmp" + str(self.sessID)
            shutil.rmtree(sessDir)
            del self
            
        except Exception as e:
            print("gaDiscrete.delete failed with error as : {}".format(e))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ga2",
    version="0.2.8",
    author="Yashas ND",
    author_email="yashasbharadwaj111@gmail.com",
    description="A simple robust Genetic Algorithm utility",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yashasbharadwaj111/GeneticAlgorithm",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Life",
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
    ],
)
        
